package cn.com.qmth.examcloud.core.oe.student.service.impl;

import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.util.JsonUtil;
import cn.com.qmth.examcloud.core.oe.common.base.Constants;
import cn.com.qmth.examcloud.core.oe.common.base.helper.ExamCacheTransferHelper;
import cn.com.qmth.examcloud.core.oe.common.base.utils.CommonUtil;
import cn.com.qmth.examcloud.core.oe.common.entity.ExamRecordDataEntity;
import cn.com.qmth.examcloud.core.oe.common.entity.FaceBiopsyEntity;
import cn.com.qmth.examcloud.core.oe.common.entity.FaceBiopsyItemEntity;
import cn.com.qmth.examcloud.core.oe.common.entity.FaceBiopsyItemStepEntity;
import cn.com.qmth.examcloud.core.oe.common.enums.ExamProperties;
import cn.com.qmth.examcloud.core.oe.common.enums.FaceBiopsyAction;
import cn.com.qmth.examcloud.core.oe.common.enums.FaceBiopsyType;
import cn.com.qmth.examcloud.core.oe.common.enums.IsSuccess;
import cn.com.qmth.examcloud.core.oe.common.repository.ExamRecordDataRepo;
import cn.com.qmth.examcloud.core.oe.common.repository.FaceBiopsyItemRepo;
import cn.com.qmth.examcloud.core.oe.common.repository.FaceBiopsyItemStepRepo;
import cn.com.qmth.examcloud.core.oe.common.repository.FaceBiopsyRepo;
import cn.com.qmth.examcloud.core.oe.student.bean.*;
import cn.com.qmth.examcloud.core.oe.student.service.ExamControlService;
import cn.com.qmth.examcloud.core.oe.student.service.ExamSessionInfoService;
import cn.com.qmth.examcloud.core.oe.student.service.FaceBiopsyService;
import cn.com.qmth.examcloud.web.bootstrap.PropertyHolder;
import cn.com.qmth.examcloud.web.helpers.GlobalHelper;
import com.google.common.collect.Lists;
import com.mysql.cj.util.StringUtils;
import org.apache.commons.lang3.RandomUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

/**
 * @Description 人脸活体检测接口实现类
 * @Author lideyin
 * @Date 2019/10/14 10:56
 * @Version 1.0
 */
@Service("faceBiopsyService")
public class FaceBiopsyServiceImpl implements FaceBiopsyService {

    @Autowired
    private FaceBiopsyRepo faceBiopsyRepo;
    @Autowired
    private FaceBiopsyItemRepo faceBiopsyItemRepo;
    @Autowired
    private FaceBiopsyItemStepRepo faceBiopsyItemStepRepo;
    @Autowired
    private ExamRecordDataRepo examRecordDataRepo;
    @Autowired
    private ExamSessionInfoService examSessionInfoService;
    @Autowired
    private ExamControlService examControlService;

    @Override
    @Transactional
    public FaceBiopsyInfo getFaceBiopsyInfo(Long rootOrgId, Long examRecordDataId, FaceBiopsyType faceBiopsyType) {
        boolean isInFreezeTime = calculateIsInFreezeTime(examRecordDataId);
        /**
         * case1.冻结时间外的校验逻辑
         * case1.1.未开启冻结时间外添加人脸活体检测的配置，则抛出异常
         * case1.2.冻结时间外已完成过1次人脸活体检测，再次调用则抛出异常
         * case1.3.冻结时间内已经存在2次失败的人脸活体检测记录，冻结时间外，再次调用抛出异常
         */
        if (!isInFreezeTime) {
            //case1.1.未开启冻结时间外添加人脸活体检测的配置，则抛出异常
            if (!isAddFaceVerifyOutFreezeTime(examRecordDataId)) {
                throw new StatusException("201004", "非冻结时间内不允许人脸活体检测");
            }
        }

        //如果是第一次进行人脸活体检测，则初始化相关信息保存并返回
        FaceBiopsyEntity faceBiopsyEntity = faceBiopsyRepo.findByExamRecordDataId(examRecordDataId);
        if (faceBiopsyEntity == null) {
            return addFirstFaceBiopsy(rootOrgId, examRecordDataId, faceBiopsyType);
        }

        //case1.2.冻结时间外已完成过1次人脸活体检测，再次调用则抛出异常
        if (!isInFreezeTime) {
            List<FaceBiopsyItemEntity> outFreezeTimeCompletedList =
                    faceBiopsyItemRepo.findByExamRecordDataIdAndCompletedAndInFreezeTime(examRecordDataId,
                            true, false);
            if (outFreezeTimeCompletedList != null && outFreezeTimeCompletedList.size() > 0) {
                throw new StatusException("201005", "冻结时间外不允许进行多次人脸活体检测");
            }

            //case1.3.冻结时间内已经存在2次失败的人脸活体检测记录，冻结时间外，再次调用抛出异常
            if (faceBiopsyEntity.getVerifiedTimes() == 2) {
                List<FaceBiopsyItemEntity> faceBiopsyItemList = faceBiopsyItemRepo.findByFaceBiopsyId(faceBiopsyEntity.getId());
                FaceBiopsyItemEntity faceBiopsyItem1 = faceBiopsyItemList.get(0);
                FaceBiopsyItemEntity faceBiopsyItem2 = faceBiopsyItemList.get(1);
                if (faceBiopsyItem1.getCompleted() && faceBiopsyItem1.getResult() == false &&
                        faceBiopsyItem2.getCompleted() && faceBiopsyItem2.getResult() == false) {
                    throw new StatusException("201006", "非法调用，活体检测已结束");
                }
            }
        }



        //如果不是第一次人脸活体检测，判断是否有未检测完的记录，
        FaceBiopsyInfo result = new FaceBiopsyInfo();
        // 未完成的活体检测列表
        List<FaceBiopsyItemEntity> unCompletedFaceBiopsyItemList =
                faceBiopsyItemRepo.findByExamRecordDataIdAndCompleted(examRecordDataId, false);

        //如果存在未完成的活体检测信息，则直接返回上次未完成的活检记录
        if (unCompletedFaceBiopsyItemList != null && !unCompletedFaceBiopsyItemList.isEmpty()) {
            FaceBiopsyItemEntity faceBiopsyItemEntity = unCompletedFaceBiopsyItemList.get(0);
            result.setFaceBiopsyItemId(faceBiopsyItemEntity.getId());
            result.setFaceVerifyMinute(getFaceBiopsyStartMinute(examRecordDataId));

            List<FaceBiopsyItemStepEntity> faceBiopsyItemStepEntityList =
                    faceBiopsyItemStepRepo.findByFaceBiopsyItemId(faceBiopsyItemEntity.getId());
            result.setVerifySteps(copyFaceBiopsyStepDomainListFrom(faceBiopsyItemStepEntityList));
            return result;
        }

        //已活检次数
        int verifiedTimes = faceBiopsyEntity.getVerifiedTimes();
        /**
         * case2.冻结时间内的校验逻辑
         * case2.1.第一次校验成功，第二次调用则抛出异常(即只有第一次失败才能再调第二次)
         * case1.2.冻结时间内活检次数不得超过2次(<=2)
         */
        if (isInFreezeTime) {
            if (verifiedTimes == 1) {
                List<FaceBiopsyItemEntity> faceBiopsyItemList = faceBiopsyItemRepo.findByFaceBiopsyId(faceBiopsyEntity.getId());
                FaceBiopsyItemEntity faceBiopsyItemEntity = faceBiopsyItemList.get(0);
                //case2.1.第一次校验成功，第二次调用则抛出异常(即只有第一次失败才能再调第二次)
                if (faceBiopsyItemEntity.getInFreezeTime() && null != faceBiopsyItemEntity.getResult()
                        && faceBiopsyItemEntity.getResult()) {
                    throw new StatusException("201006", "非法请求，冻结时间内无需多次活体检测");
                }
            } else {
                throw new StatusException("201006", "非法请求，冻结时间内最多只能两次活检");
            }
        }

        //没有超过，则追加活检次数
        verifiedTimes++;
        return appendFaceBiopsy(faceBiopsyEntity.getId(), verifiedTimes, examRecordDataId, faceBiopsyType);
    }

    @Override
    @Transactional
    public SaveFaceBiopsyResultResp saveFaceBiopsyResult(SaveFaceBiopsyResultReq req) {
        //构建业务实体
        SaveFaceBiopsyResultResp resp = buildSaveFaceBiopsyResultResp(req.getExamRecordDataId(), req.getVerifySteps());

        //更新人脸活体检测结果至数据库
        updateFaceBiopsyResult(req.getExamRecordDataId(), req.getFaceBiopsyItemId(), req.getVerifySteps(),
                resp.getVerifyResult(), resp.getErrorMessage());

        //同步更新考试记录表中的活体检测结果
        IsSuccess faceVerifyResult = resp.getVerifyResult() ? IsSuccess.SUCCESS : IsSuccess.FAILED;
        examRecordDataRepo.updateFaceVerifyResult(faceVerifyResult, req.getExamRecordDataId());

        return resp;
    }

    /**
     * 获取人脸活体检测开始分钟数
     *
     * @param examRecordDataId 考试记录id
     * @return Integer
     */
    @Override
    public Integer getFaceBiopsyStartMinute(Long examRecordDataId) {
        //case1.1.第一次调用，返回冻结时间内或冻结时间外下次活检时间
        FaceBiopsyEntity faceBiopsyEntity = faceBiopsyRepo.findByExamRecordDataId(examRecordDataId);
        //如果是一次活检都没有做过，则认为是第一次
        if (faceBiopsyEntity == null) {
            return calculateFaceBiopsyStartMinute(examRecordDataId);
        }

        boolean isInFreezeTime = calculateIsInFreezeTime(examRecordDataId);
        /**
         * case1.冻结时间内的逻辑
         * case1.1.第一次调用，返回冻结时间内下次活检时间
         * case1.2.第二次调用，
         * case1.2.1.如果第一次活检失败，
         * case1.2.1.1.如果第二次活检未完成，重新计算下次活检时间
         * case1.2.1.2.如果第二次活检失败，返回null
         * case1.2.1.3.如果第二次活检成功，返回冻结时间外活检开始的时间
         * case1.2.2.如果第一次活检成功，
         * case1.2.2.1.但未开启冻结时间外追加人脸活检的配置，返回null
         * case1.2.2.2.且开启了冻结时间外追加人脸活检的配置，返回冻结时间外活检开始的时间
         *
         */
        if (isInFreezeTime) {
            if (faceBiopsyEntity.getVerifiedTimes() <= 2) {
                List<FaceBiopsyItemEntity> faceBiopsyItemList = faceBiopsyItemRepo.findByFaceBiopsyId(faceBiopsyEntity.getId());
                FaceBiopsyItemEntity faceBiopsyItemEntity = faceBiopsyItemList.get(0);
                //第一次活体检测成功，第二次调用时
                if (faceBiopsyItemEntity.getInFreezeTime() && faceBiopsyItemEntity.getCompleted()
                        && faceBiopsyItemEntity.getResult()) {
                    //如果未开启冻结时间外人脸活体检测，则直接返回null
                    if (!isAddFaceVerifyOutFreezeTime(examRecordDataId)) {
                        return null;
                    }

                    //如果开启冻结时间外人脸活体检测，则直接返回冻结时间外的活检开始时间
                    return calculateFaceBiopsyStartMinute(examRecordDataId, true,
                            FreezeTimeBorder.OUT);
                }

                //如果进行了两次活检
                if (faceBiopsyEntity.getVerifiedTimes() == 2) {
                    faceBiopsyItemEntity = faceBiopsyItemList.get(1);
                    //如果进行了两次活检，且第二次活检失败，冻结时间内再次调用，返回null
                    if (faceBiopsyItemEntity.getInFreezeTime() && faceBiopsyItemEntity.getCompleted()
                            && faceBiopsyItemEntity.getResult() == false) {
                        return null;
                    }

                    //如果进行了两次活检，且第二次活检成功，冻结时间内再次调用，返回冻结时间外的值
                    if (faceBiopsyItemEntity.getInFreezeTime() && faceBiopsyItemEntity.getCompleted()
                            && faceBiopsyItemEntity.getResult()) {
                        return calculateFaceBiopsyStartMinute(examRecordDataId, true,
                                FreezeTimeBorder.OUT);
                    }
                }
            }
        }

        /**
         * case2.冻结时间外的校验逻辑,
         * case2.1.冻结时间外已完成过1次人脸活体检测，再次调用则返回null
         * case2.2.冻结时间内两次活检均失败，冻结时间外再次调用则返回null
         */
        if (!isInFreezeTime) {
            List<FaceBiopsyItemEntity> outFreezeTimeCompletedList =
                    faceBiopsyItemRepo.findByExamRecordDataIdAndCompletedAndInFreezeTime(examRecordDataId,
                            true, false);
            //冻结时间外已完成过1次人脸活体检测，再次调用则返回null
            if (outFreezeTimeCompletedList != null && outFreezeTimeCompletedList.size() > 0) {
                return null;
            }

            //冻结时间内两次活检均失败，冻结时间外再次调用则返回null
            if (faceBiopsyEntity.getVerifiedTimes() == 2) {
                List<FaceBiopsyItemEntity> faceBiopsyItemList = faceBiopsyItemRepo.findByFaceBiopsyId(faceBiopsyEntity.getId());
                FaceBiopsyItemEntity faceBiopsyItem1 = faceBiopsyItemList.get(0);
                FaceBiopsyItemEntity faceBiopsyItem2 = faceBiopsyItemList.get(1);
                if (faceBiopsyItem1.getCompleted() && faceBiopsyItem1.getResult() == false &&
                        faceBiopsyItem2.getCompleted() && faceBiopsyItem2.getResult() == false) {
                    return null;
                }
            }
        }

        return calculateFaceBiopsyStartMinute(examRecordDataId);
    }

    /**
     * 构建保存人脸检测结果的业务实体
     *
     * @param examRecordDataId
     * @param verifySteps
     * @return
     */
    private SaveFaceBiopsyResultResp buildSaveFaceBiopsyResultResp(Long examRecordDataId,
                                                                   List<FaceBiopsyStepInfo> verifySteps) {
        //本次检测的最终结果是否成功(检测步骤中只要有一项检测失败，则认为检测失败)
        boolean finalIsSuccess = true;
        String errorMsg = null;
        /**
         * 是否结束考试
         * case1:人脸比对失败(即活检第1步,action==FACE_COMPARE，照片非本人或检测中多人脸)时，结束考试
         * case2:冻结时间内，第2次人脸活检整体失败，结束考试
         * case3:冻结时间外，人脸活检失败，结束考试
         */
        boolean isEndExam = false;
        //检测次数
        int verifyTimes = getVerifyTimes(examRecordDataId);
        //是否在冻结时间内
        boolean isInFreezeTime = calculateIsInFreezeTime(examRecordDataId);

        // 检测结果至少有一条检测结果不允许为空
        if (!verifySteps.stream().anyMatch(p -> null != p.getResult())) {
            throw new StatusException("201005", "检测结果不允许为空");
        }
        for (FaceBiopsyStepInfo stepInfo : verifySteps) {
            //如果检测结果为空，则直接跳过
            if (stepInfo.getResult() == null) {
                continue;
            }

            if (!stepInfo.getResult()) {
                errorMsg = stepInfo.getErrorMsg();
            }

            //如果当前步骤为超时，则直接跳出循环
            if ((stepInfo.getReadTimeout() != null && stepInfo.getReadTimeout())) {
                errorMsg = "未在指定时间内完成指令";
                finalIsSuccess = false;
                break;
            }

            //整个检测步骤中只要有一次失败或超时，则认为整个活检失败
            if (!stepInfo.getResult()) {
                finalIsSuccess = false;
            }

            switch (stepInfo.getAction()) {
                case FACE_COMPARE:
                    //如果第一步检测部分失败(照片非本人或检测中多人脸)，需要结束考试
                    Map faceCompareResult = JsonUtil.fromJson(stepInfo.getResultJson(), Map.class);
                    //TODO 取照片检测结果
                    Boolean isStranger = Boolean.valueOf(faceCompareResult.get("isStranger").toString());
                    Boolean isPass = Boolean.valueOf(faceCompareResult.get("isPass").toString());
                    Boolean existsSystemError = Boolean.valueOf(faceCompareResult.get("existsSystemError").toString());
                    //case1.1.有陌生人(即多人脸)，不管是否比对成功，直接结束考试
                    if (isStranger) {
                        isEndExam = true;
                    }
                    //case1.2.无陌生人且检测失败且不是系统错误(即照片非本人)，也直接结束考试
                    else {
                        if (!isPass && !existsSystemError) {
                            isEndExam = true;
                        }
                    }
                    break;
                case HAPPY:
                case SERIOUS:
                    //如果此前步骤已判断为需要结束考试则不作任何处理
                    if (!isEndExam) {
                        if (isInFreezeTime) {
                            //如果冻结时间内第2次检测失败，则需要结束考试
                            if (verifyTimes > 1 && !stepInfo.getResult()) {
                                isEndExam = true;
                            }
                        } else {
                            //冻结时间外检测失败，直接结束考试
                            if (!stepInfo.getResult()) {
                                isEndExam = true;
                            }
                        }
                    }
                    break;
            }

        }
        //构建业务实体
        SaveFaceBiopsyResultResp resp = new SaveFaceBiopsyResultResp();
        resp.setEndExam(isEndExam);
        resp.setNeedNextVerify(calculateNeedNextVerify(
                isEndExam, isInFreezeTime, finalIsSuccess, verifyTimes, examRecordDataId));
        resp.setVerifyResult(finalIsSuccess);
        resp.setErrorMessage(errorMsg);
        return resp;
    }

    /**
     * 更新人脸活体检测结果至数据库
     *
     * @param examRecordDataId
     * @param faceBiopsyItemId
     * @param verifySteps
     * @param finalIsSuccess
     * @param errorMsg
     */
    public void updateFaceBiopsyResult(Long examRecordDataId, Long faceBiopsyItemId, List<FaceBiopsyStepInfo> verifySteps,
                                       boolean finalIsSuccess, String errorMsg) {
        List<FaceBiopsyItemStepEntity> faceBiopsyItemStepEntityList =
                faceBiopsyItemStepRepo.findByFaceBiopsyItemId(faceBiopsyItemId);
        for (FaceBiopsyStepInfo stepInfo : verifySteps) {
            //循环更新新活体检测每步的执行结果
            FaceBiopsyItemStepEntity faceBiopsyItemStepEntity = faceBiopsyItemStepEntityList.stream().
                    filter(p -> p.getId().equals(stepInfo.getStepId())).findFirst().get();
            faceBiopsyItemStepEntity.setErrorMsg(stepInfo.getErrorMsg());
            faceBiopsyItemStepEntity.setResourceType(stepInfo.getResourceType());
            faceBiopsyItemStepEntity.setResourceRelativePath(getRelativePath(stepInfo.getResourceUrl()));
            faceBiopsyItemStepEntity.setExt1(stepInfo.getResultJson());//存储动作执行结果
            faceBiopsyItemStepEntity.setResult(stepInfo.getResult());
        }
        //批量更新活检步骤执行结果
        faceBiopsyItemStepRepo.saveAll(faceBiopsyItemStepEntityList);
        //更新人脸活体检测最终结果
        faceBiopsyItemRepo.updateFaceBiopsyItemResult(faceBiopsyItemId, finalIsSuccess, errorMsg, true);
        faceBiopsyRepo.updateFaceBiopsyResult(examRecordDataId, finalIsSuccess, errorMsg);
    }

    /**
     * 计算是否需要下次活检
     *
     * @param isEndExam
     * @param isInFreezeTime
     * @param finalIsSuccess
     * @param verifyTimes
     * @return
     */
    private boolean calculateNeedNextVerify(boolean isEndExam, boolean isInFreezeTime, boolean finalIsSuccess,
                                            int verifyTimes, Long examRecordDataId) {
        //考试已结束，不需要继续下次活检
        if (isEndExam) {
            return false;
        }

        //如果考试未结束的情况
        if (isInFreezeTime) {
            if (verifyTimes == 1) {
                //冻结时间内第一次活检失败，则需要下次活检
                if (!finalIsSuccess) {
                    return true;
                }
            } else {
                //冻结时间内非第一次活检失败，则不需要下次活检
                if (!finalIsSuccess) {
                    return false;
                }
            }

            //冻结时间内第一次或第二次活检成功，且开启冻结时间外增加一次活检，则需要下次活检，否则不需要
            ExamRecordDataEntity examRecordData = GlobalHelper.getEntity(examRecordDataRepo, examRecordDataId,
                    ExamRecordDataEntity.class);
            return isAddFaceVerifyOutFreezeTime(examRecordData.getExamId(),
                    examRecordData.getOrgId(), examRecordData.getStudentId());
        } else {
            //冻结时间外，无论活检成功或失败，均不可进行下次活检
            return false;
        }
    }

    /**
     * 获取活体检测次数
     *
     * @param examRecordDataId
     * @return
     */
    private int getVerifyTimes(Long examRecordDataId) {
        FaceBiopsyEntity faceBiopsy = faceBiopsyRepo.findByExamRecordDataId(examRecordDataId);
        return faceBiopsy.getVerifiedTimes();
    }

    /**
     * 根据完整路径获取相对路径
     *
     * @param resourceUrl
     * @return
     */
    private String getRelativePath(String resourceUrl) {
        if (!StringUtils.isNullOrEmpty(resourceUrl)) {
            String domain = PropertyHolder.getString("$upyun.site.1.domain");
            String backupDomain = PropertyHolder.getString("$upyun.site.1.domain.backup");
            if (!StringUtils.isNullOrEmpty(domain)) {
                resourceUrl = resourceUrl.replace(domain, "");
            }
            if (!StringUtils.isNullOrEmpty(backupDomain)) {
                resourceUrl = resourceUrl.replace(backupDomain, "");
            }
        }
        return resourceUrl;
    }

    /**
     * 第一次添加人脸活体检测结果
     *
     * @param rootOrgId
     * @param examRecordDataId
     * @param faceBiopsyType
     * @return
     */
    public FaceBiopsyInfo addFirstFaceBiopsy(Long rootOrgId, Long examRecordDataId, FaceBiopsyType faceBiopsyType) {
        //保存人脸活体检测相关信息
        Long faceBiopsyId = addFaceBiopsyEntity(rootOrgId, examRecordDataId);
        Long faceBiopsyItemId = addFaceBiopsyItemEntity(examRecordDataId, faceBiopsyType, faceBiopsyId);
        List<FaceBiopsyItemStepEntity> faceBiopsyItemStepEntityList =
                addFaceBiopsyItemStepList(examRecordDataId, faceBiopsyItemId);

        //构建业务实体
        return buildFaceBiopsyInfo(examRecordDataId, faceBiopsyItemId, faceBiopsyItemStepEntityList);
    }

    /**
     * 追加人脸活体检测结果
     *
     * @param faceBiopsyId
     * @param verifiedTimes
     * @param examRecordDataId
     * @param faceBiopsyType
     * @return
     */
    @Transactional
    public FaceBiopsyInfo appendFaceBiopsy(Long faceBiopsyId, Integer verifiedTimes, Long examRecordDataId,
                                           FaceBiopsyType faceBiopsyType) {
        //更新人脸活体检测次数
        FaceBiopsyEntity faceBiopsyEntity = GlobalHelper.getEntity(faceBiopsyRepo, faceBiopsyId, FaceBiopsyEntity.class);
        faceBiopsyEntity.setVerifiedTimes(verifiedTimes);
        faceBiopsyRepo.save(faceBiopsyEntity);
        //添加人脸活体检测明细
        Long faceBiopsyItemId = addFaceBiopsyItemEntity(examRecordDataId, faceBiopsyType, faceBiopsyId);
        //添加人脸活体检测步骤
        List<FaceBiopsyItemStepEntity> faceBiopsyItemStepEntityList =
                addFaceBiopsyItemStepList(examRecordDataId, faceBiopsyItemId);

        return buildFaceBiopsyInfo(examRecordDataId, faceBiopsyItemId, faceBiopsyItemStepEntityList);

    }

    /**
     * 构建人脸检测基本信息的业务实体对象
     *
     * @param examRecordDataId
     * @param faceBiopsyItemId
     * @param faceBiopsyItemStepEntityList
     * @return
     */
    private FaceBiopsyInfo buildFaceBiopsyInfo(Long examRecordDataId, Long faceBiopsyItemId,
                                               List<FaceBiopsyItemStepEntity> faceBiopsyItemStepEntityList) {
        //构建业务实体
        FaceBiopsyInfo faceBiopsyInfo = new FaceBiopsyInfo();
        faceBiopsyInfo.setFaceBiopsyItemId(faceBiopsyItemId);
        faceBiopsyInfo.setFaceVerifyMinute(getFaceBiopsyStartMinute(examRecordDataId));
        faceBiopsyInfo.setVerifySteps(copyFaceBiopsyStepDomainListFrom(faceBiopsyItemStepEntityList));
        return faceBiopsyInfo;
    }

    private List<FaceBiopsyStepInfo> copyFaceBiopsyStepDomainListFrom(List<FaceBiopsyItemStepEntity> faceBiopsyItemStepEntityList) {
        List<FaceBiopsyStepInfo> resultList = new ArrayList<>();
        for (FaceBiopsyItemStepEntity entity : faceBiopsyItemStepEntityList) {
            FaceBiopsyStepInfo domain = new FaceBiopsyStepInfo();
            domain.setAction(entity.getAction());
            domain.setStay(entity.getActionStay());
            domain.setStepId(entity.getId());
            resultList.add(domain);
        }
        return resultList;
    }

    /**
     * 添加活体检测的具体步骤
     *
     * @param examRecordDataId 考试记录id
     * @param faceBiopsyItemId 人脸活体检测明细id
     */
    private List<FaceBiopsyItemStepEntity> addFaceBiopsyItemStepList(Long examRecordDataId, Long faceBiopsyItemId) {
        List<FaceBiopsyItemStepEntity> faceBiopsyItemStepEntityList = new ArrayList<FaceBiopsyItemStepEntity>();

        //step1.人脸比对
        FaceBiopsyItemStepEntity firstStep = new FaceBiopsyItemStepEntity();
        firstStep.setExamRecordDataId(examRecordDataId);
        firstStep.setFaceBiopsyItemId(faceBiopsyItemId);
        firstStep.setAction(FaceBiopsyAction.FACE_COMPARE);
        faceBiopsyItemStepEntityList.add(firstStep);

        //生成乱序表情集合
        List<FaceBiopsyAction> shuffledExpressionList = generateShuffleExpressions();

        //step2随机表情
        faceBiopsyItemStepEntityList.add(
                randomGenerateExpressionStep(examRecordDataId, faceBiopsyItemId, shuffledExpressionList));

        //step3随机表情
        faceBiopsyItemStepEntityList.add(
                randomGenerateExpressionStep(examRecordDataId, faceBiopsyItemId, shuffledExpressionList));

        faceBiopsyItemStepRepo.saveAll(faceBiopsyItemStepEntityList);
        return faceBiopsyItemStepEntityList;
    }

    /**
     * 乱序生成人脸表情
     *
     * @return
     */
    private List<FaceBiopsyAction> generateShuffleExpressions() {
        List<FaceBiopsyAction> expressionList = Lists.newArrayList();
        expressionList.add(FaceBiopsyAction.HAPPY);
        expressionList.add(FaceBiopsyAction.SERIOUS);
        Collections.shuffle(expressionList);
        return expressionList;
    }

    /**
     * 随机生成表情相关步骤
     *
     * @param examRecordDataId
     * @param faceBiopsyItemId
     * @return
     */
    private FaceBiopsyItemStepEntity randomGenerateExpressionStep(Long examRecordDataId,
                                                                  Long faceBiopsyItemId,
                                                                  List<FaceBiopsyAction> shuffledExpressionList) {
        FaceBiopsyItemStepEntity secondStep = new FaceBiopsyItemStepEntity();
        secondStep.setExamRecordDataId(examRecordDataId);
        secondStep.setFaceBiopsyItemId(faceBiopsyItemId);
        secondStep.setAction(takeShuffledExpression(shuffledExpressionList));//获取一条随机生成表情指令
        secondStep.setActionStay(randomGenerateActionStay());//随机生成指令时长(3至6秒)
        return secondStep;
    }

    /**
     * 获取一条随机生成的指令
     *
     * @param shuffledExpressionList
     * @return
     */
    private FaceBiopsyAction takeShuffledExpression(List<FaceBiopsyAction> shuffledExpressionList) {
        //取完第一条并删除
        FaceBiopsyAction expression = shuffledExpressionList.get(0);
        shuffledExpressionList.remove(0);
        return expression;
    }


    /**
     * 随机生成指令持续的时长
     *
     * @return
     */
    private Integer randomGenerateActionStay() {
        int minStay = PropertyHolder.getInt("oe.faceBiopsy.minActionStay", 3);
        int maxStay = PropertyHolder.getInt("oe.faceBiopsy.maxActionStay", 6);
        return RandomUtils.nextInt(minStay, maxStay);
    }

    /**
     * 添加人脸活体检测明细
     *
     * @param examRecordDataId 考试记录id
     * @param faceBiopsyType   人脸活体检测类型
     * @param faceBiopsyId     人脸活体检测id
     * @return Long 人脸活体检测明细id
     */
    private Long addFaceBiopsyItemEntity(Long examRecordDataId, FaceBiopsyType faceBiopsyType,
                                         Long faceBiopsyId) {
        FaceBiopsyItemEntity faceBiopsyItemEntity = new FaceBiopsyItemEntity();
        faceBiopsyItemEntity.setExamRecordDataId(examRecordDataId);
        faceBiopsyItemEntity.setFaceBiopsyId(faceBiopsyId);
        faceBiopsyItemEntity.setFaceBiopsyType(faceBiopsyType);
        faceBiopsyItemEntity.setCompleted(false);
        faceBiopsyItemEntity.setInFreezeTime(calculateIsInFreezeTime(examRecordDataId));
        faceBiopsyItemRepo.save(faceBiopsyItemEntity);
        return faceBiopsyItemEntity.getId();
    }

    /**
     * 添加人脸活体检测结果
     *
     * @param rootOrgId        学校id
     * @param examRecordDataId 考试记录id
     * @return Long 人脸活体检测结果id
     */
    private Long addFaceBiopsyEntity(Long rootOrgId, Long examRecordDataId) {
        FaceBiopsyEntity faceBiopsyEntity = new FaceBiopsyEntity();
        faceBiopsyEntity.setRootOrgId(rootOrgId);
        faceBiopsyEntity.setExamRecordDataId(examRecordDataId);
        faceBiopsyEntity.setVerifiedTimes(1);
        faceBiopsyRepo.save(faceBiopsyEntity);
        return faceBiopsyEntity.getId();
    }

    /**
     * @param examRecordDataId 考试记录id
     * @return
     */
    private Integer calculateFaceBiopsyStartMinute(Long examRecordDataId) {
        return calculateFaceBiopsyStartMinute(examRecordDataId, false, null);
    }

    /**
     * 计算人脸活体检测开始分钟数
     *
     * @param examRecordDataId            考试记录id
     * @param isSpecifiedFreezeTimeBorder 是否指定冻结时间
     * @param freezeTimeBorder            冻结时间边界
     * @return
     */
    private Integer calculateFaceBiopsyStartMinute(Long examRecordDataId, Boolean isSpecifiedFreezeTimeBorder,
                                                   FreezeTimeBorder freezeTimeBorder) {
        ExamRecordDataEntity examRecordData = GlobalHelper.getEntity(examRecordDataRepo, examRecordDataId,
                ExamRecordDataEntity.class);
        Long examId = examRecordData.getExamId();
        Long orgId = examRecordData.getOrgId();
        Long studentId = examRecordData.getStudentId();

        //再次校验是否开启人脸活体校验
        String isFaceVerifyStr = ExamCacheTransferHelper.getCachedExamProperty(examId, orgId,
                studentId, ExamProperties.IS_FACE_VERIFY.name()).getValue();

        //如果未开启人脸活体检测，则返回null
        if (!Constants.isTrue.equals(isFaceVerifyStr)) {
            return null;
        }

        ExamSessionInfo examSessionInfo = examSessionInfoService.getExamSessionInfo(studentId);
        if (examSessionInfo == null) {
            throw new StatusException("201002", "考试会话已过期");
        }

        Integer heartbeat = examSessionInfo.getHeartbeat();
        Integer freezeTime = examSessionInfo.getFreezeTime();

        //如果特殊指定了获取冻结时间外或冻结时间内的计算方法，则根据参数分别调用相关方法
        if (isSpecifiedFreezeTimeBorder) {
            if (freezeTimeBorder == FreezeTimeBorder.IN) {
                return generateInFreezeTimeFaceBiopsyStartMinute(examRecordDataId, examId, orgId, studentId, heartbeat);
            } else {
                return generateOutFreezeTimeFaceBiopsyStartMinute(examId, orgId, studentId, heartbeat, freezeTime);
            }
        }

        //未特殊指定则自动判断，如果冻结时间内，使用冻结时间内的计算方式
        if (calculateIsInFreezeTime(examRecordDataId)) {
            return generateInFreezeTimeFaceBiopsyStartMinute(examRecordDataId, examId, orgId, studentId, heartbeat);
        }
        //如果超过冻结时间则使用冻结时间外的计算方式
        else {
            return generateOutFreezeTimeFaceBiopsyStartMinute(examId, orgId, studentId, heartbeat, freezeTime);
        }
    }

    /**
     * 冻结时间边界枚举
     */
    enum FreezeTimeBorder {
        /**
         * 冻结时间内
         */
        IN,
        /**
         * 冻结时间外
         */
        OUT
    }

    /**
     * 冻结时间内生成人脸活体检测的开始时间
     *
     * @param examId
     * @param orgId
     * @param heartbeat
     * @return
     */
    private Integer generateInFreezeTimeFaceBiopsyStartMinute(Long examRecordDataId, Long examId, Long orgId, Long studentId,
                                                              Integer heartbeat) {
        Integer verifyTimes = 0;
        //第一次调用，返回冻结时间内或冻结时间外下次活检时间
        FaceBiopsyEntity faceBiopsyEntity = faceBiopsyRepo.findByExamRecordDataId(examRecordDataId);
        if (faceBiopsyEntity == null) {
            verifyTimes = 1;
        } else {
            verifyTimes = faceBiopsyEntity.getVerifiedTimes();
        }

        //如果没有进行过人脸检测(即第一次活体检测)
        if (verifyTimes == 1) {
            String faceVerifyStartMinuteStr = ExamCacheTransferHelper.getCachedExamProperty(examId, orgId,
                    studentId, ExamProperties.FACE_VERIFY_START_MINUTE.name()).getValue();
            Integer faceVerifyStartMinute = Integer.valueOf(faceVerifyStartMinuteStr);

            String faceVerifyEndMinuteStr = ExamCacheTransferHelper.getCachedExamProperty(examId, orgId,
                    studentId, ExamProperties.FACE_VERIFY_END_MINUTE.name()).getValue();
            Integer faceVerifyEndMinute = Integer.valueOf(faceVerifyEndMinuteStr);
            //	case1.如果考生已使用的考试时间(即心跳时间)还未达到系统设置的活体检测开始时间，
            //	则实际活体检测时间=random(配置结束时间-配置开始时间)-考试已用时间
            if (heartbeat < faceVerifyStartMinute) {
                return CommonUtil.calculationRandomNumber(faceVerifyStartMinute, faceVerifyEndMinute) - heartbeat;
            }
            //	case2如果配置开始时间<考生已使用的考试时间<配置结束时间，
            //	则实际活体检测时间=random(配置结束时间-考试已用时间)-考试已用时间，如果结果小于1分钟则默认1分钟
            else if (heartbeat >= faceVerifyStartMinute && heartbeat < faceVerifyEndMinute) {
                int verifyTime = CommonUtil.calculationRandomNumber(heartbeat, faceVerifyEndMinute) - heartbeat;
                return verifyTime < 1 ? 1 : verifyTime;
            }
            //case3如果考试已用时间>配置结束时间，则默认1分钟后开始人脸检测
            else {
                return 1;
            }
        }
        //如果是第二次人脸
        else {
            return CommonUtil.calculationRandomNumber(
                    PropertyHolder.getInt("oe.faceBiopsy.minSecondFaceCheckMinute", 1),
                    PropertyHolder.getInt("oe.faceBiopsy.maxSecondFaceCheckMinute", 4));
        }
    }

    /**
     * 冻结时间外生成人脸活体检测的开始时间
     *
     * @return
     */
    private Integer generateOutFreezeTimeFaceBiopsyStartMinute(Long examId, Long orgId,
                                                               Long studentId, Integer heartbeat, Integer freezeTime) {
        //如果冻结时间外不添加活体检测，则直接返回null
        if (!isAddFaceVerifyOutFreezeTime(examId, orgId, studentId)) {
            return null;
        }

        String strOutFreezeTimeFaceVerifyStartMinute = ExamCacheTransferHelper.getCachedExamProperty(examId, orgId,
                studentId, ExamProperties.OUT_FREEZE_TIME_FACE_VERIFY_START_MINUTE.name()).getValue();
        Integer faceVerifyStartMinute = Integer.valueOf(strOutFreezeTimeFaceVerifyStartMinute);

        String strOutFreezeTimeFaceVerifyEndMinute = ExamCacheTransferHelper.getCachedExamProperty(examId, orgId,
                studentId, ExamProperties.OUT_FREEZE_TIME_FACE_VERIFY_END_MINUTE.name()).getValue();
        Integer faceVerifyEndMinute = Integer.valueOf(strOutFreezeTimeFaceVerifyEndMinute);

        //活检开始时间 = (冻结时间 + random(配置开始时间,配置结束时间)) -  考试已用时间
        int result = freezeTime + CommonUtil.calculationRandomNumber(faceVerifyStartMinute, faceVerifyEndMinute) - heartbeat;
        return result < 1 ? 1 : result;
    }

    /**
     * 是否允许冻结时间外添加人脸活体检测
     *
     * @param examId
     * @param orgId
     * @return
     */
    private boolean isAddFaceVerifyOutFreezeTime(Long examId, Long orgId, Long studentId) {
        String addFaceVerifyOutFreezeTime = ExamCacheTransferHelper.getCachedExamProperty(examId, orgId,
                studentId, ExamProperties.ADD_FACE_VERIFY_OUT_FREEZE_TIME.name()).getValue();
        return Constants.isTrue.equals(addFaceVerifyOutFreezeTime);
    }

    /**
     * 是否允许冻结时间外添加人脸活体检测
     *
     * @param examRecordDataId
     * @return
     */
    private boolean isAddFaceVerifyOutFreezeTime(Long examRecordDataId) {
        ExamRecordDataEntity examRecordData = GlobalHelper.getEntity(examRecordDataRepo, examRecordDataId,
                ExamRecordDataEntity.class);
        return isAddFaceVerifyOutFreezeTime(examRecordData.getExamId(),
                examRecordData.getOrgId(), examRecordData.getStudentId());
    }

    /**
     * 当前考试是否在冻结时间内
     *
     * @param examRecordDataId 考试记录id
     * @return
     */
    private boolean calculateIsInFreezeTime(Long examRecordDataId) {
        ExamRecordDataEntity examRecordData = GlobalHelper.getEntity(examRecordDataRepo, examRecordDataId,
                ExamRecordDataEntity.class);

        ExamSessionInfo examSessionInfo = examSessionInfoService.getExamSessionInfo(examRecordData.getStudentId());
        if (examSessionInfo == null) {
            throw new StatusException("201003", "考试会话已过期");
        }
        Long usedMilliseconds = examControlService.calculateExamUsedMilliseconds(examSessionInfo);

        return usedMilliseconds < examSessionInfo.getFreezeTime() * 60 * 1000;
    }
}
